﻿#region EK.Math, Copyright(C) 2003-2004 Eran Kampf, Licensed under LGPL.
//	EK.Math math library
//	Copyright (C) 2003-2004  
//	Eran Kampf
//	eran@ekampf.com
//	http://www.ekampf.com
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#endregion
#region Using directives

using System;
using System.Diagnostics;
using System.Collections.Generic;
using System.Text;

#endregion

namespace Matrix4PerfTest.Generic
{
	#region IKernelMatrix4 interface
	public interface IKernelMatrix4<T> 
		where T : struct
	{
		Matrix4<T> Add(Matrix4<T> a, Matrix4<T> b);
		void Add(Matrix4<T> a, Matrix4<T> b, Matrix4<T> result);

		Matrix4<T> Multiply(Matrix4<T> a, Matrix4<T> b);
		void Multiply(Matrix4<T> a, Matrix4<T> b, Matrix4<T> result);

		Matrix4<T> Transpose(Matrix4<T> matrix);
	}
	#endregion

	#region C# IKernelMatrix4 Kernel
	public struct KernelMatrix4F : IKernelMatrix4<float>
	{
		#region IKernelMatrix4<float> Members

		public Matrix4<float> Add(Matrix4<float> a, Matrix4<float> b)
		{
			return new Matrix4<float>(
				a.M11 + b.M11, a.M12 + b.M12, a.M13 + b.M13, a.M14 + b.M14,
				a.M21 + b.M21, a.M22 + b.M22, a.M23 + b.M23, a.M24 + b.M24,
				a.M31 + b.M31, a.M32 + b.M32, a.M33 + b.M33, a.M34 + b.M34,
				a.M41 + b.M41, a.M42 + b.M42, a.M43 + b.M43, a.M44 + b.M44
				);
		}

		public void Add(Matrix4<float> a, Matrix4<float> b, Matrix4<float> result)
		{
			result.M11 = a.M11 + b.M11;
			result.M12 = a.M12 + b.M12;
			result.M13 = a.M13 + b.M13;
			result.M14 = a.M14 + b.M14;

			result.M21 = a.M21 + b.M21;
			result.M22 = a.M22 + b.M22;
			result.M23 = a.M23 + b.M23;
			result.M24 = a.M24 + b.M24;

			result.M31 = a.M31 + b.M31;
			result.M32 = a.M32 + b.M32;
			result.M33 = a.M33 + b.M33;
			result.M34 = a.M34 + b.M34;

			result.M41 = a.M41 + b.M41;
			result.M42 = a.M42 + b.M42;
			result.M43 = a.M43 + b.M43;
			result.M44 = a.M44 + b.M44;
		}

		public Matrix4<float> Multiply(Matrix4<float> a, Matrix4<float> b)
		{
			return new Matrix4<float>(
				a.M11 * b.M11 + a.M12 * b.M21 + a.M13 * b.M31 + a.M14 * b.M41,
				a.M11 * b.M12 + a.M12 * b.M22 + a.M13 * b.M32 + a.M14 * b.M42,
				a.M11 * b.M13 + a.M12 * b.M23 + a.M13 * b.M33 + a.M14 * b.M43,
				a.M11 * b.M14 + a.M12 * b.M24 + a.M13 * b.M34 + a.M14 * b.M44,

				a.M21 * b.M11 + a.M22 * b.M21 + a.M23 * b.M31 + a.M24 * b.M41,
				a.M21 * b.M12 + a.M22 * b.M22 + a.M23 * b.M32 + a.M24 * b.M42,
				a.M21 * b.M13 + a.M22 * b.M23 + a.M23 * b.M33 + a.M24 * b.M43,
				a.M21 * b.M14 + a.M22 * b.M24 + a.M23 * b.M34 + a.M24 * b.M44,

				a.M31 * b.M11 + a.M32 * b.M21 + a.M33 * b.M31 + a.M34 * b.M41,
				a.M31 * b.M12 + a.M32 * b.M22 + a.M33 * b.M32 + a.M34 * b.M42,
				a.M31 * b.M13 + a.M32 * b.M23 + a.M33 * b.M33 + a.M34 * b.M43,
				a.M31 * b.M14 + a.M32 * b.M24 + a.M33 * b.M34 + a.M34 * b.M44,

				a.M41 * b.M11 + a.M42 * b.M21 + a.M43 * b.M31 + a.M44 * b.M41,
				a.M41 * b.M12 + a.M42 * b.M22 + a.M43 * b.M32 + a.M44 * b.M42,
				a.M41 * b.M13 + a.M42 * b.M23 + a.M43 * b.M33 + a.M44 * b.M43,
				a.M41 * b.M14 + a.M42 * b.M24 + a.M43 * b.M34 + a.M44 * b.M44
				);
		}

		public void Multiply(Matrix4<float> a, Matrix4<float> b, Matrix4<float> result)
		{
			result.M11 = a.M11 * b.M11 + a.M12 * b.M21 + a.M13 * b.M31 + a.M14 * b.M41;
			result.M12 = a.M11 * b.M12 + a.M12 * b.M22 + a.M13 * b.M32 + a.M14 * b.M42;
			result.M13 = a.M11 * b.M13 + a.M12 * b.M23 + a.M13 * b.M33 + a.M14 * b.M43;
			result.M14 = a.M11 * b.M14 + a.M12 * b.M24 + a.M13 * b.M34 + a.M14 * b.M44;

			result.M21 = a.M21 * b.M11 + a.M22 * b.M21 + a.M23 * b.M31 + a.M24 * b.M41;
			result.M22 = a.M21 * b.M12 + a.M22 * b.M22 + a.M23 * b.M32 + a.M24 * b.M42;
			result.M23 = a.M21 * b.M13 + a.M22 * b.M23 + a.M23 * b.M33 + a.M24 * b.M43;
			result.M24 = a.M21 * b.M14 + a.M22 * b.M24 + a.M23 * b.M34 + a.M24 * b.M44;

			result.M31 = a.M31 * b.M11 + a.M32 * b.M21 + a.M33 * b.M31 + a.M34 * b.M41;
			result.M32 = a.M31 * b.M12 + a.M32 * b.M22 + a.M33 * b.M32 + a.M34 * b.M42;
			result.M33 = a.M31 * b.M13 + a.M32 * b.M23 + a.M33 * b.M33 + a.M34 * b.M43;
			result.M34 = a.M31 * b.M14 + a.M32 * b.M24 + a.M33 * b.M34 + a.M34 * b.M44;

			result.M41 = a.M41 * b.M11 + a.M42 * b.M21 + a.M43 * b.M31 + a.M44 * b.M41;
			result.M42 = a.M41 * b.M12 + a.M42 * b.M22 + a.M43 * b.M32 + a.M44 * b.M42;
			result.M43 = a.M41 * b.M13 + a.M42 * b.M23 + a.M43 * b.M33 + a.M44 * b.M43;
			result.M44 = a.M41 * b.M14 + a.M42 * b.M24 + a.M43 * b.M34 + a.M44 * b.M44;
		}

		public Matrix4<float> Transpose(Matrix4<float> matrix)
		{
			Matrix4<float> result = new Matrix4<float>();

			result.M12 = matrix.M21;
			result.M13 = matrix.M31;
			result.M14 = matrix.M41;
			result.M23 = matrix.M32;
			result.M24 = matrix.M42;
			result.M34 = matrix.M43;

			result.M21 = matrix.M12;
			result.M31 = matrix.M13;
			result.M41 = matrix.M14;
			result.M32 = matrix.M23;
			result.M42 = matrix.M24;
			result.M43 = matrix.M34;

			return result;
		}

		#endregion
	}
	#endregion

	#region DirectX IKernelMatrix4 Kernel
	public class KerneDirectXlMatrix4 : IKernelMatrix4<float>
	{
		private Microsoft.DirectX.Matrix ToDirectX(Matrix4<float> value)
		{
			Microsoft.DirectX.Matrix m = new Microsoft.DirectX.Matrix();
			// Since directx use column-major matrices we transpose our matrix.
			m.M11 = value.M11; m.M12 = value.M21; m.M13 = value.M31; m.M14 = value.M41;
			m.M21 = value.M12; m.M22 = value.M22; m.M23 = value.M32; m.M24 = value.M42;
			m.M31 = value.M13; m.M32 = value.M23; m.M33 = value.M33; m.M34 = value.M43;
			m.M41 = value.M14; m.M42 = value.M24; m.M43 = value.M34; m.M44 = value.M44;

			return m;
		}
		public Matrix4<float> FromDirectX(Microsoft.DirectX.Matrix value)
		{
			Matrix4<float> m = new Matrix4<float>();
			// Since directx use column-major matrices we transpose the matrix.
			m.M11 = value.M11; m.M12 = value.M21; m.M13 = value.M31; m.M14 = value.M41;
			m.M21 = value.M12; m.M22 = value.M22; m.M23 = value.M32; m.M24 = value.M42;
			m.M31 = value.M13; m.M32 = value.M23; m.M33 = value.M33; m.M34 = value.M43;
			m.M41 = value.M14; m.M42 = value.M24; m.M43 = value.M34; m.M44 = value.M44;

			return m;
		}

		#region IKernelMatrix4<float> Members

		public Matrix4<float> Add(Matrix4<float> a, Matrix4<float> b)
		{
			throw new NotImplementedException();
		}

		public void Add(Matrix4<float> a, Matrix4<float> b, Matrix4<float> result)
		{
			throw new NotImplementedException();
		}

		public Matrix4<float> Multiply(Matrix4<float> a, Matrix4<float> b)
		{
			Microsoft.DirectX.Matrix dxA = ToDirectX(a);
			Microsoft.DirectX.Matrix dxB = ToDirectX(b);

			return FromDirectX(Microsoft.DirectX.Matrix.Multiply(dxA, dxB));
		}

		public void Multiply(Matrix4<float> a, Matrix4<float> b, Matrix4<float> result)
		{
			throw new NotImplementedException();
		}

		public Matrix4<float> Transpose(Matrix4<float> matrix)
		{
			throw new NotImplementedException();
		}

		#endregion
	}
	#endregion

	/// <summary>
	/// Represents a 2 dimentional matrix
	/// </summary>
	[Serializable]
	public struct Matrix4<T> : ICloneable
		where T : struct
	{
		#region Private Fields
		private T _m11, _m12, _m13, _m14;
		private T _m21, _m22, _m23, _m24;
		private T _m31, _m32, _m33, _m34;
		private T _m41, _m42, _m43, _m44;
		#endregion

		#region Constructors
		public Matrix4(
			T m11, T m12, T m13, T m14,
			T m21, T m22, T m23, T m24,
			T m31, T m32, T m33, T m34,
			T m41, T m42, T m43, T m44
			)
		{
			_m11 = m11; _m12 = m12; _m13 = m13; _m14 = m14;
			_m21 = m21; _m22 = m22; _m23 = m23; _m24 = m24;
			_m31 = m31; _m32 = m32; _m33 = m33; _m34 = m34;
			_m41 = m41; _m42 = m42; _m43 = m43; _m44 = m44;
		}
		public Matrix4(List<T> elements)
		{
			Debug.Assert(elements != null);
			Debug.Assert(elements.Count >= 16);

			_m11 = elements[0]; _m12 = elements[1]; _m13 = elements[2]; _m14 = elements[3];
			_m21 = elements[4]; _m22 = elements[5]; _m23 = elements[6]; _m24 = elements[7];
			_m31 = elements[8]; _m32 = elements[9]; _m33 = elements[10]; _m34 = elements[11];
			_m41 = elements[12]; _m42 = elements[13]; _m43 = elements[14]; _m44 = elements[15];
		}
		public Matrix4(Matrix4<T> m)
		{
			_m11 = m.M11; _m12 = m.M12; _m13 = m.M13; _m14 = m.M14;
			_m21 = m.M21; _m22 = m.M22; _m23 = m.M23; _m24 = m.M24;
			_m31 = m.M31; _m32 = m.M32; _m33 = m.M33; _m34 = m.M34;
			_m41 = m.M41; _m42 = m.M42; _m43 = m.M43; _m44 = m.M44;
		}
		#endregion

		#region ICloneable Members
		/// <summary>
		/// Creates an exact copy of this <see cref="Matrix4"/> object.
		/// </summary>
		/// <returns>The <see cref="Matrix4"/> object this method creates, cast as an object.</returns>
		object ICloneable.Clone()
		{
			return new Matrix4<T>(this);
		}
		/// <summary>
		/// Creates an exact copy of this <see cref="Matrix4"/> object.
		/// </summary>
		/// <returns>The <see cref="Matrix4"/> object this method creates.</returns>
		public Matrix4<T> Clone()
		{
			return new Matrix4<T>(this);
		}
		#endregion

		#region Public Properties
		public T M11
		{
			get { return _m11; }
			set { _m11 = value; }
		}
		public T M12
		{
			get { return _m12; }
			set { _m12 = value; }
		}
		public T M13
		{
			get { return _m13; }
			set { _m13 = value; }
		}
		public T M14
		{
			get { return _m14; }
			set { _m14 = value; }
		}

		public T M21
		{
			get { return _m21; }
			set { _m21 = value; }
		}
		public T M22
		{
			get { return _m22; }
			set { _m22 = value; }
		}
		public T M23
		{
			get { return _m23; }
			set { _m23 = value; }
		}
		public T M24
		{
			get { return _m24; }
			set { _m24 = value; }
		}

		public T M31
		{
			get { return _m31; }
			set { _m31 = value; }
		}
		public T M32
		{
			get { return _m32; }
			set { _m32 = value; }
		}
		public T M33
		{
			get { return _m33; }
			set { _m33 = value; }
		}
		public T M34
		{
			get { return _m34; }
			set { _m34 = value; }
		}

		public T M41
		{
			get { return _m41; }
			set { _m41 = value; }
		}
		public T M42
		{
			get { return _m42; }
			set { _m42 = value; }
		}
		public T M43
		{
			get { return _m43; }
			set { _m43 = value; }
		}
		public T M44
		{
			get { return _m44; }
			set { _m44 = value; }
		}
		#endregion

		#region System.Object Overrides
		public override int GetHashCode()
		{
			return
				_m11.GetHashCode() ^ _m12.GetHashCode() ^
				_m21.GetHashCode() ^ _m22.GetHashCode();
		}
		#endregion

		public static Matrix4<T> operator *(Matrix4<T> a, Matrix4<T> b)
		{
			return Kernel.Multiply(a, b);
		}

		public static void Swap(ref T a, ref T b)
		{
			T t = a; a = b; b = t;
		}

		public void Transpose()
		{
			Matrix4<T>.Swap(ref _m12, ref _m21);
		}

		private static IKernelMatrix4<T> kernel;
		public static IKernelMatrix4<T> Kernel
		{
			get
			{
				return kernel;
			}
			set
			{
				kernel = value;
			}
		}
	}
}
